#include "TH1F.h"
#include "TProfile.h"
#include "TMath.h"
#include "TF1.h"
#include "TLegend.h"
#include "TCanvas.h"
 //
#include "TROOT.h"
//#include "TStyle.h"
#include "TTree.h"
#include "TFile.h"
#include "TGraphErrors.h"
#include "TStyle.h"
#include <string>
#include <cmath>
#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>
//

using namespace std;

const int debug = 0; 
//const int debug = 3; //DS test
//const bool regeneratePlots = true;
const bool regeneratePlots = false;

const int kNChan = 32;
const int kMaxLineLength = kNChan * 6; // max 4 digits + 1 space per channel + 1 extra for margin

const int kNSkip = 100; // skip some first samples

const int kNBinsPerSweep = 1000; // 0.2 us per bin  
const int kNSweep = 36;   

const int kNPadSweepX = 4;
const int kNPadSweepY = 2;
const int kNPlotSweep = kNPadSweepX * kNPadSweepY;
const int kSelectedChanSweep[kNPlotSweep] = { 0, 1, 2, 3, 28, 29, 30, 31};

int analyse_file(const int iChip, const char *infile, const char *outfile, const char *config)
{
  gStyle->SetOptStat(0);
  gStyle->SetOptFit(0111);
  gStyle->SetStatX(0.5); 

  FILE* fout = fopen(outfile, "w");
  fprintf(fout, "%9s %12s\n", "Channel #", " Ped      RMS");

  char outdir[50];
  char command[100];

  if (regeneratePlots) {
    char chipVer[10];
    if (iChip<2000) {
      sprintf(chipVer, "V2");
    }
    else if (iChip>=2000 && iChip<6000) {
      sprintf(chipVer, "V3");
    }
    else if (iChip>=6000) {
      sprintf(chipVer, "V4");
    }

    sprintf(outdir, "png/calib/%s/%d", chipVer, iChip);
    sprintf(command, ".! mkdir -p %s", outdir);
    gROOT->ProcessLine(command);
    cout << " executing command " << command << endl;
  }

  ifstream inADC;
  inADC.open(infile);
  cout << " reading file " << infile << endl;
    
  // counter stuff
  int ichan = 0;
  int isweep = 0;
  int ibin = 0; // in sweep

  char hname[100];
  char hdesc[100];
  TH1F* hSample[32];
  TH1F* hSweep[32][kNSweep];

  for (ichan=0; ichan<32; ichan++) {
    sprintf(hname, "hSample_%d", ichan);
    sprintf(hdesc, "All ADC values - Chan %d", ichan);
    hSample[ichan] = new TH1F(hname, hdesc, 1024, -0.5, 1023.5);

    for (isweep=0; isweep<kNSweep; isweep++) {
      sprintf(hname, "hSweep_%d_%d", ichan, isweep);
      sprintf(hdesc, "TimeSweep %d (%d bins) - Chan %d", isweep, kNSweep, ichan);
      hSweep[ichan][isweep] = new TH1F(hname, hdesc, 
				       kNBinsPerSweep, -0.5, kNBinsPerSweep - 0.5);
    }
  }
  TProfile* hBase = new TProfile("hBase", "hBase", 32, -0.5, 31.5);
  
  int nlines  = 0; // over whole file

  // sample loop, inside block of interesting values, between zeros
  int nsamples = 0;
  char line[kMaxLineLength];
  int samples[kNChan];

  while ( inADC.good() ) {
    inADC.getline(line, kMaxLineLength);
    if (! inADC.good()) break;
    nlines++;      

    // split line into the expected 32 pieces 
    int ns = 0;
    char * pch;
    pch = strtok (line," ");
    while (pch != NULL) {
      // printf ("ns %d pch %s\n",ns, pch);
      samples[ns] = atoi(pch);
      pch = strtok (NULL, " ");
      ns++;
    }
    if (debug>2) { cout << "nlines " << nlines << " ns " << ns << endl; }

    if (nlines > kNSkip) {
      if (debug>2) { cout << "Fill hSample nlines " << nlines  << endl; }
      for (ichan=0; ichan<32; ichan++) {
	hSample[ichan]->Fill(samples[ichan]);
	hBase->Fill(ichan, samples[ichan]);
      }
      int iline = nlines - kNSkip - 1;
      isweep = iline/kNBinsPerSweep;
      ibin = iline % kNBinsPerSweep;
      if (isweep < kNSweep) {
	for (ichan=0; ichan<32; ichan++) {
	  hSweep[ichan][isweep]->SetBinContent(ibin+1, samples[ichan]);
	}
      }
    }
  } // inADC

  for (ichan=0; ichan<32; ichan++) { 
    hSample[ichan]->GetXaxis()->SetRangeUser(hBase->GetBinContent(ichan+1) - 10,
					     hBase->GetBinContent(ichan+1) + 10 );
    fprintf(fout, "%2d%7s %5.1f     %3.2f\n", ichan, "", 
	    hSample[ichan]->GetMean(), hSample[ichan]->GetRMS() );
  }

  if (regeneratePlots) {
    // make plots
    // fit channel info, and store fit parameters (with errors) and chi2 values
    char name[50], description[100];
    TCanvas* cBase = new TCanvas("cBase", "Baseline vs Chan #", 900, 600);
    TCanvas* cSample = new TCanvas("cSample", "Sample distribution", 1800, 1200);
    TCanvas* cSweep = new TCanvas("cSweep", "Swep distribution", 1800, 1200);
    
    cSample->Divide(8,4);
    cSweep->Divide(kNPadSweepX, kNPadSweepY);
    
    cBase->cd();
    hBase->Draw();
    sprintf(name, "%s/base_chip_%d_%s.png", outdir, iChip, config);
    cBase->SaveAs(name);
    
    for (ichan=0; ichan<32; ichan++) { 
      hSample[ichan]->SetMarkerStyle(20);
      cSample->cd(ichan + 1);
      hSample[ichan]->GetXaxis()->SetRangeUser(hBase->GetBinContent(ichan+1) - 10,
					       hBase->GetBinContent(ichan+1) + 10 );
      
      sprintf(description, "Sample distr for chip %d chan %d", iChip, ichan);
      hSample[ichan]->GetXaxis()->SetTitle("ADC");
      hSample[ichan]->GetYaxis()->SetTitle("Yield");
      hSample[ichan]->SetTitle(description);
      hSample[ichan]->Draw();
    } // ichan
    cSample->Modified();
    sprintf(name, "%s/sample_chip_%d_%s.png", outdir, iChip, config);
    cSample->SaveAs(name);
    
    for (isweep=0; isweep<32; isweep++) {
      int isel = 0;
      for (ichan=0; ichan<32; ichan++) { 
	if (kSelectedChanSweep[isel] == ichan) {
	  hSweep[ichan][isweep]->SetMarkerStyle(20);
	  isel++;

	  cSweep->cd(isel);    
	  sprintf(description, "Sweep %d for chip %d chan %d", isweep, iChip, ichan);
	  hSweep[ichan][isweep]->GetYaxis()->SetTitle("ADC");
	  hSweep[ichan][isweep]->GetXaxis()->SetTitle("Timebin");
	  hSweep[ichan][isweep]->SetTitle(description);
	  hSweep[ichan][isweep]->Draw();
	}
      } // ichan
      cSweep->Modified();
      sprintf(name, "%s/sweep_%d_chip_%d_%s.png", outdir, isweep, iChip, config);
      cSweep->SaveAs(name);
    }
  } // plots

  fclose(fout);
  return 0;
}

int main (int argc, char *argv[])
{
  if (argc<4) {
    printf("not enough arguments: argc = %d\nSyntax should be:\n./noise.exe chipstring config adcfilename\n",argc);
    printf("Example:\n./noise.exe 100000 30mV_noise /home/tpc/robot/20180725/123510_sampa100000/123510_sampa100000_trorc00_link00_30mV_noise_adc.txt\n",argc);
    return -1;
  }
  int iChip = atoi(argv[1]);

  char* infile = argv[3];
  // find dirname (last /)
  string str(infile);
  size_t found;
  found=str.find_last_of("/");
  char* outfile = Form("%s/noiseAna_sampa%s_%s.log",  str.substr(0,found).c_str(), argv[1], argv[2]);

  int iret = analyse_file(iChip, infile, outfile, argv[2]);
  return iret;
}
