#include <iostream>
#include <string>
#include <cstdlib>
#include <boost/program_options.hpp>
#include <boost/tokenizer.hpp>

//#include "trorc.hpp"
#include "gbt_link.hpp"
#include "gbt_sca_comm.hpp"
#include "gbt_sca.hpp"
#include "gbt_sca_i2c.hpp"
#include "sampa.hpp"
//#include "fec.hpp"

// Namespaces
using namespace gbt;
namespace bpo = boost::program_options;

// Globals
bool verbose = false;
bool broadcast = false;

int main( int argc, char** argv )
{
  uint32_t fec_idx;
  // Program options
  bpo::variables_map vm;
  bpo::options_description ped_opts("Pedestal Options", 1224);
  bpo::options_description reg_opts("SAMPA Register Options", 1224);
  bpo::options_description misc_opts("Miscellaneous Options", 1224);
  bpo::options_description opt_all;

  // Miscellaneous options
  misc_opts.add_options()
    ("help,h", "Print help message")
    ("verbose,v", bpo::bool_switch(&verbose), "Be verbose")
    //  ("sampa,s", bpo::value<std::vector<std::string>>()->multitoken(), "Define SAMPA device address and GBT channel")
    //  ("sampa,s", bpo::value<std::string>()->multitoken(), "Define SAMPA device address and GBT channel (chipID:GBTchannel)")
    ("sampa,s", bpo::value<std::string>(), "Define SAMPA device address and SCA I2C channel (chipID:I2Cchannel)")
    //  ("mask,m", bpo::value<std::string>()->default_value("0x1"), "FEC mask")
    ("idx,i", bpo::value<uint32_t>(&fec_idx)->default_value(0), "GBTx0 index (0, 2, 4, ...)")
    ("broadcast,b", bpo::bool_switch(&broadcast), "Broadcast write to all channels")
    ;

  // Pedestal register options
  ped_opts.add_options()
    ("ped-rd", bpo::value<std::vector<std::string>>()->multitoken(), "Read pedestal memory (--ped-rd [channel] [pedaddr])")
    ("ped-wr", bpo::value<std::vector<std::string>>()->multitoken(), "Write to pedestal memory (--ped-wr [channel] [pedaddr] [val])")
    ;
  // Other register options
  reg_opts.add_options()
    ("g-rd", bpo::value<std::string>(), "Read global register (--g-rd [pedaddr])")
    ("g-wr", bpo::value<std::vector<std::string>>()->multitoken(), "Write to global register (--g-wr [pedaddr] [val])")
    ("rd", bpo::value<std::vector<std::string>>()->multitoken(), "Read channel register (--rd [channel] [addr])")
    ("wr", bpo::value<std::vector<std::string>>()->multitoken(), "Write to channel register (--wr [channel] [addr] [val])")
    ;

  opt_all.add(misc_opts).add(reg_opts).add(ped_opts);

  // Check for parser errors
  bool parseErr = false;
  try {
    bpo::store(bpo::parse_command_line(argc, argv, opt_all), vm);
    bpo::notify(vm);
  } catch (bpo::error& e) {
    std::cerr << "Error: " << e.what() << std::endl;
    parseErr = true;
  } if ( parseErr || vm.count("help") || argc==1) {
    std::cout << opt_all << std::endl;
    if ( parseErr ) return 64;
    else return 0;
  }


  // Start doing stuff
  int device_number(0);
  int bar_number(1);

  std::unique_ptr<trorc::Device> trorc;
  std::unique_ptr<trorc::Bar> bar;
  std::unique_ptr<ScaI2cSampa> i2c_c;
  try {
    trorc.reset(new trorc::Device(device_number));
    bar.reset(new trorc::Bar(*trorc, bar_number));
  } catch (int e) {
    std::cerr << "ERROR: Failed to intialize T-RORC -- " << librorc::errMsg(e)
              << std::endl;
    exit(1);
  }

  // Parsing 
  try {
   
    /* Tokenize SAMPA list
     * Can be of the form chipID:GBT_CH; chip1,chip2,chip3:ch1,ch2,ch3; chip1-chip3:ch1-ch3; 
     * */
    bool comma_list = false;
    bool dash_range = false;
    std::vector<uint32_t> chip_ids;
    std::vector<uint32_t> gbt_channels;

    if( !vm["sampa"].empty() ) {
      
      // Check for colons
      boost::char_separator<char> colon(":");
      boost::tokenizer <boost::char_separator<char>> toks(vm["sampa"].as<std::string>(), colon);
      std::vector<std::string> token_list;
      for(auto token : toks) { token_list.push_back(token); }

      // Some counting
      size_t n_comma_l  = std::count(token_list[0].begin(), token_list[0].end(), ',');
      size_t n_dash_l   = std::count(token_list[0].begin(), token_list[0].end(), '-');
      size_t n_comma_r = std::count(token_list[1].begin(), token_list[1].end(), ',');
      size_t n_dash_r  = std::count(token_list[1].begin(), token_list[1].end(), '-');

      // Check tokens (chip ID):(gbt channel)
      boost::char_separator<char> dash("-");
      boost::char_separator<char> comma(",");
      boost::tokenizer <boost::char_separator<char>> chip_id_toks(token_list[0], (n_comma_l > 0) ? comma : dash);
      boost::tokenizer <boost::char_separator<char>> gbt_ch_toks(token_list[1], (n_comma_r > 0) ? comma : dash);
      // Extract chip ID tokens
      for(auto token : chip_id_toks) { chip_ids.push_back(stoi(token, nullptr, 0)); } 
      // Extract channel tokens
      for(auto token : gbt_ch_toks) { gbt_channels.push_back(stoi(token, nullptr, 0)); }
      // Count number of tokens
      //uint32_t n_chip_id = std::distance(chip_id_toks.begin(), chip_id_toks.end());
      //uint32_t n_ch = std::distance(gbt_ch_toks.begin(), gbt_ch_toks.end());

      // Comma separated list
      if( (n_comma_l > 0) && (n_comma_l == n_comma_r) ) {
        comma_list = true; // XXX: Actually never used in code...
      } 
      // Ranged list
      else if( (n_dash_l == 1) && (n_dash_r == 1) ) {
        if( std::abs(chip_ids.back()-chip_ids.front())==std::abs(gbt_channels.back()-gbt_channels.front()) ) {
          dash_range = true;
        } else {
	  std::cout << "Chip ID and GBT channel ranges are not equal" << std::endl;
          exit(1);
        }
      }
      // Single SAMPA
      else if ( (chip_ids.size() == 1) && (gbt_channels.size() == 1) ) {
        comma_list = true; // XXX: Sort of
      }
      else {
	std::cout << "Sampa definitions are not recognized" << std::endl;
      }
    }
    
    // FEC mask
    //uint32_t fec_mask = stoi(vm["mask"].as<std::string>(), nullptr, 0);
    //    for() {
    // Open link
    //int fec_idx = 0; // TODO: Implement FEC index option 
    GscTrorcHdlcSw sca_comm(*bar, fec_idx, true, false);
    // If ranged: m=gbt_channels
    uint32_t m = 0;
    if(dash_range) m = gbt_channels.front();
    for(uint32_t n = (dash_range?chip_ids.front():0); (dash_range?(n<=chip_ids.back()) : (n<chip_ids.size())); (dash_range?(n++,m++):n++) ) {
      i2c_c.reset(new ScaI2cSampa(sca_comm, (dash_range?m:gbt_channels.at(n)), (dash_range?n:chip_ids.at(n))));
      Sampa sampa(*i2c_c);
      /* Pedestal memories */
      // Read
      if( !vm["ped-rd"].empty() && (vm["ped-rd"].as<std::vector<std::string>>()).size() == 2 ) {
	uint8_t  channel = (uint8_t)  stoi(vm["ped-rd"].as<std::vector<std::string>>()[0], nullptr, 0);
	uint16_t addr    = (uint16_t) stoi(vm["ped-rd"].as<std::vector<std::string>>()[1], nullptr, 0);
	uint16_t val     = (uint16_t) sampa.readPedestal(channel, addr);
	// Output
	if( verbose ) {
	  std::cout << "I2C Addr: 0x" << std::hex << std::setfill('0') << std::setw(2)
		    << std::uppercase << (uint32_t) n 
		    << ", SCA CH" << std::uppercase << (uint32_t) m+3 << std::endl;
	} //else {
	std::cout << "0x" << std::hex << std::setfill('0') << std::setw(3)
		  << std::uppercase << (uint32_t) addr << ": 0x" 
		  << std::hex << std::setfill('0') << std::setw(3)
		  << std::uppercase << (uint32_t) val 
		  << std::endl;
	//}
      }
      // Write
      if( !vm["ped-wr"].empty() && ( (vm["ped-wr"].as<std::vector<std::string>>()).size() > 0 ) ) {
	int n_args = (vm["ped-wr"].as<std::vector<std::string>>()).size();
          uint8_t  channel = broadcast ? 0xBC
	    : (uint8_t)  stoi(vm["ped-wr"].as<std::vector<std::string>>()[0], nullptr, 0);
          uint16_t addr    = (uint16_t) stoi(vm["ped-wr"].as<std::vector<std::string>>()[((broadcast&&n_args==2)?0:1)], nullptr, 0);
          uint16_t val     = (uint16_t) stoi(vm["ped-wr"].as<std::vector<std::string>>()[((broadcast&&n_args==2)?1:2)], nullptr, 0);
          sampa.writePedestal(channel, addr, val);
      }
        
      /* Global registers */
      // Read
      if(!vm["g-rd"].empty()) {
	uint16_t addr = (uint16_t) stoi(vm["g-rd"].as<std::string>(), nullptr, 0);
	uint8_t val = sampa.readGlobalReg(addr);
	// Output
	if( verbose ) {
	  std::cout << "I2C Addr: 0x" << std::hex << std::setfill('0') << std::setw(2)
		    << std::uppercase << (uint32_t) n 
		    << ", SCA CH" << std::uppercase << (uint32_t) m+3 << std::endl;
	} //else {
	std::cout << "0x" << std::hex << std::setfill('0') << std::setw(3)
		  << std::uppercase << (uint32_t) addr << ": 0x" 
		  << std::hex << std::setfill('0') << std::setw(3)
		  << std::uppercase << (uint32_t) val 
		  << std::endl;
      }
      // Write
      if(!vm["g-wr"].empty() && ( (vm["g-wr"].as<std::vector<std::string>>()).size() == 2 )) {
	uint16_t addr = (uint16_t) stoi(vm["g-wr"].as<std::vector<std::string>>()[0], nullptr, 0);
	uint8_t val  = (uint8_t) stoi(vm["g-wr"].as<std::vector<std::string>>()[1], nullptr, 0);
	sampa.writeGlobalReg(addr, val);
      }

      /* Channel registers */
      // Read
      if(!vm["rd"].empty() && ( (vm["rd"].as<std::vector<std::string>>()).size() == 2 )) {
	uint16_t addr = (uint16_t) stoi(vm["rd"].as<std::vector<std::string>>()[0], nullptr, 0);
	uint8_t  channel = (uint8_t) stoi(vm["rd"].as<std::vector<std::string>>()[0], nullptr, 0);
	uint16_t val = sampa.readChannelReg(channel, addr);
	// Output
	if( verbose ) {
	  std::cout << "I2C Addr: 0x" << std::hex << std::setfill('0') << std::setw(2)
		    << std::uppercase << (uint32_t) n 
		    << ", SCA CH" << std::uppercase << (uint32_t) m+3 << std::endl;
	} //else {
	std::cout << "0x" << std::hex << std::setfill('0') << std::setw(3)
		  << std::uppercase << (uint32_t) addr << ": 0x" 
		  << std::hex << std::setfill('0') << std::setw(3)
		  << std::uppercase << (uint32_t) val 
		  << std::endl;

      }
      // Write
      if( !vm["wr"].empty() && ( (vm["wr"].as<std::vector<std::string>>()).size() > 0 ) ) {
	int n_args = (vm["wr"].as<std::vector<std::string>>()).size();
          uint8_t  channel = broadcast ? 0xBC
	    : (uint8_t)  stoi(vm["wr"].as<std::vector<std::string>>()[0], nullptr, 0);
          uint16_t addr    = (uint16_t) stoi(vm["wr"].as<std::vector<std::string>>()[((broadcast&&n_args==2)?0:1)], nullptr, 0);
          uint16_t val     = (uint16_t) stoi(vm["wr"].as<std::vector<std::string>>()[((broadcast&&n_args==2)?1:2)], nullptr, 0);
          sampa.writeChannelReg(channel, addr, val);
      }
    } // end for loop SAMPA
    //} // end for loop FEC
  } catch (std::exception& e) {
    std::cerr << e.what() << ", exiting" << std::endl;
    exit(100);
  }


  return 0;
}
