-- ------------------------------------------------------------------------------
-- Lightweight HDLC communication for use with GBT
--   Receives 64-bit HDLC frame payload and assembles complete HDLC frame
--   Outputs bit-stuffed HDLC frame data at 2 bits/cycle
--
-- kirsch@fias.uni-frankfurt.de
-- ------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity gbt_hdlc_light is
  port (
    -- clk             : in std_logic;    -- FIXME: (temporary) port to two clock domains for
    -- rst             : in std_logic;    --          use in current G-RORC

    tx_clk : in std_logic;
    tx_rst : in std_logic;

    rx_clk : in std_logic;
    rx_rst : in std_logic;

    hdlc_tx_payload : in  std_logic_vector(63 downto 0);
    hdlc_tx_start   : in  std_logic;                      -- Start Tx (implicitly enables reception of one frame)
    hdlc_rx_start   : in  std_logic;                      -- Start Rx (explicity enable reception of one frame)
    hdlc_tx_mode    : in  std_logic_vector(1 downto 0);   -- Regular packet (00), Svl CONNECT (01), Svl RESET (10), Svl TEST (11)
    hdlc_tx_addr    : in  std_logic_vector(7 downto 0);   -- Only for regular packets
    hdlc_tx_rdy     : out std_logic;
    hdlc_tx_seq_nr  : out std_logic_vector(7 downto 0);
    hdlc_tx_fsm     : out std_logic_vector(3 downto 0);

    hdlc_rx_payload : out std_logic_vector(63 downto 0);  -- Rx HDLC payload
    hdlc_rx_addr    : out std_logic_vector(7 downto 0);   -- Rx HDLC address field
    hdlc_rx_ctrl    : out std_logic_vector(7 downto 0);   -- Rx HDLC control field
    hdlc_rx_rdy     : out std_logic;                      -- Frame received
    hdlc_rx_svl_cmd : out std_logic;                      -- Last frame was supervisory level command
    hdlc_rx_crc     : out std_logic_vector(31 downto 0);
    hdlc_rx_seq_nr  : out std_logic_vector(7 downto 0);
    hdlc_rx_fsm     : out std_logic_vector(3 downto 0);

    rx_2b           : in  std_logic_vector(1 downto 0);
    tx_2b           : out std_logic_vector(1 downto 0)
  );
end gbt_hdlc_light;

architecture arc of gbt_hdlc_light is

  --
  -- Calculate CRC for shift by 2
  function next_crc16_d2 (
     crc:  std_logic_vector(15 downto 0);
     data: std_logic_vector(1 downto 0)
    )  return std_logic_vector is
      variable d:      std_logic_vector(1 downto 0);
      variable c:      std_logic_vector(15 downto 0);
      variable newcrc: std_logic_vector(15 downto 0);
    begin
      d := data;
      c := crc;
      newcrc(0)  := c(2);
      newcrc(1)  := c(3);
      newcrc(2)  := c(4) xor c(0) xor d(0);
      newcrc(3)  := c(5) xor c(1) xor d(1);
      newcrc(4)  := c(6);
      newcrc(5)  := c(7);
      newcrc(6)  := c(8);
      newcrc(7)  := c(9);
      newcrc(8)  := c(10);
      newcrc(9)  := c(11) xor c(0) xor d(0);
      newcrc(10) := c(12) xor c(1) xor d(1);
      newcrc(11) := c(13);
      newcrc(12) := c(14);
      newcrc(13) := c(15);
      newcrc(14) := c(0) xor d(0);
      newcrc(15) := c(1) xor d(1);
      return newcrc;
  end next_crc16_d2;

  --
  -- Calculate CRC for shift by 8
  function next_crc16_d8 (
    crc  : std_logic_vector(15 downto 0);
    data : std_logic_vector(7 downto 0)
  ) return std_logic_vector is
    variable newcrc : std_logic_vector(15 downto 0);
  begin
    newcrc := next_crc16_d2(next_crc16_d2(next_crc16_d2(next_crc16_d2(crc,
                data(1 downto 0)), data(3 downto 2)), data(5 downto 4)), data(7 downto 6));
    return newcrc;
  end next_crc16_d8;

  --
  -- Constants & signals
  constant HDLC_ABORT      : std_logic_vector(7 downto 0) := x"7f";
  constant HDLC_SOF        : std_logic_vector(7 downto 0) := x"7e";
  alias    HDLC_EOF        : std_logic_vector(7 downto 0) is HDLC_SOF;
  constant HDLC_ABORT_CYC  : natural := HDLC_ABORT'length/2;
  constant HDLC_SOF_CYC    : natural := HDLC_SOF'length/2;
  alias    HDLC_EOF_CYC    : natural is HDLC_SOF_CYC;

  constant C_RX_SEQ_NR_INC : positive := 1;
  constant C_TX_SEQ_NR_INC : positive := 1;

  constant TX_MODE_REGULAR     : std_logic_vector(1 downto 0) := "00";
  constant TX_MODE_SVL_CONNECT : std_logic_vector(1 downto 0) := "01";
  constant TX_MODE_SVL_RESET   : std_logic_vector(1 downto 0) := "10";
  constant TX_MODE_SVL_TEST    : std_logic_vector(1 downto 0) := "11";
  constant TX_CTRL_SVL_CONNECT : std_logic_vector(7 downto 0) := x"2f";
  constant TX_CTRL_SVL_RESET   : std_logic_vector(7 downto 0) := x"8f";
  constant TX_CTRL_SVL_TEST    : std_logic_vector(7 downto 0) := x"e3";

  signal   hdlc_tx_start_re       : std_logic;
  signal   hdlc_tx_start_r        : std_logic;
  signal   hdlc_rx_start_re       : std_logic;
  signal   hdlc_rx_start_r        : std_logic;

  --
  -- Tx stuff
  signal   tx_hdlc_addr     : std_logic_vector(7 downto 0) := (others => '0');
  signal   tx_hdlc_ctrl     : std_logic_vector(7 downto 0) := (others => '0');
  signal   tx_hdlc_payload  : std_logic_vector(63 downto 0) := (others => '0');
  constant tx_hdlc_zero_pad : std_logic_vector(31 downto 0) := (others => '0');
  signal   tx_hdlc_fcs      : std_logic_vector(15 downto 0) := (others => '0');

  constant N_BIT_FRAME_LOW : natural := tx_hdlc_addr'length + tx_hdlc_ctrl'length + 
                                        tx_hdlc_payload'length + tx_hdlc_zero_pad'length;

  signal   tx_frame_low     : std_logic_vector(N_BIT_FRAME_LOW - 1 downto 0);
  signal   tx_frame_c       : std_logic_vector(N_BIT_FRAME_LOW - 1 downto 0);
  signal   tx_frame_cap     : std_logic;
  signal   tx_frame_shift8  : std_logic;
  signal   tx_hdlc_fcs_c    : std_logic_vector(7 downto 0);

  signal   tx_shift_reg     : std_logic_vector(9 downto 0);
  signal   tx_shift_reg_c   : std_logic_vector(7 downto 0);

  type state_send_type is (SS_RESET, SS_IDLE, SS_SHIFT_SOF, SS_SHIFT_FRAME_LOW, SS_SHIFT_FCS, SS_SHIFT_EOF);
  signal   cstate_tx  : state_send_type;
  signal   nstate_tx  : state_send_type;

  signal   tx_2b_i          : std_logic_vector(tx_2b'range);

  signal   tx_request       : std_logic;
  signal   tx_mode          : std_logic_vector(hdlc_tx_mode'range);
  signal   tx_addr          : std_logic_vector(hdlc_tx_addr'range);
  signal   tx_done          : std_logic;
  signal   tx_shift         : std_logic;
  signal   tx_bcnt          : unsigned(7 downto 0);
  signal   tx_bcnt_sval     : unsigned(tx_bcnt'range);
  signal   tx_bcnt_dec      : std_logic;
  signal   tx_bcnt_inh      : std_logic;
  signal   tx_seq_nr        : unsigned(7 downto 0) := (others => '0');
  signal   tx_history       : std_logic_vector(4 downto 0) := (others => '0');
  signal   tx_history_4co   : std_logic;
  signal   tx_history_5co   : std_logic;

  signal   rx_seq_nr        : unsigned(7 downto 0) := (others => '0');

  type tx_load_type is (LOAD_NONE, LOAD_ABORT, LOAD_FRAME_SOF, LOAD_FRAME_EOF, LOAD_FRAME_LOW, LOAD_FRAME_FCS);
  signal tx_shift_load_sel  : tx_load_type;
  signal tx_shift_load      : std_logic;

  signal   tx_crc_clr       : std_logic := '0';

  signal   bts_enable       : std_logic := '0';
  signal   bts_enable_r     : std_logic := '0';
  signal   bts_flag         : std_logic := '0';
  signal   bts_select       : std_logic_vector(2 downto 0);
  signal   bts_req          : std_logic_vector(1 downto 0);
  alias    bts_req_even     : std_logic is bts_req(0);
  alias    bts_req_odd      : std_logic is bts_req(1);

  --
  -- Rx stuff
  signal   rx_hdlc_addr       : std_logic_vector(7 downto 0) := (others => '0');
  signal   rx_hdlc_ctrl       : std_logic_vector(7 downto 0) := (others => '0');
  signal   rx_hdlc_data       : std_logic_vector(63 downto 0) := (others => '0');
  signal   rx_hdlc_fcs        : std_logic_vector(15 downto 0) := (others => '0');
  signal   rx_hdlc_fcs_c      : std_logic_vector(15 downto 0) := (others => '0');
  signal   rx_hdlc_fcs_c_rbo  : std_logic_vector(15 downto 0) := (others => '0');
  signal   rx_hdlc_fcs_recalc : std_logic_vector(15 downto 0) := (others => '0');
  signal   rx_shift_reg       : std_logic_vector(24 downto 0);

  signal   rx_sof_det         : std_logic_vector(1 downto 0);
  signal   rx_abort_det       : std_logic_vector(1 downto 0);
  signal   rx_dbts_req        : std_logic_vector(1 downto 0);
  signal   rx_dbts_req_r      : std_logic_vector(rx_dbts_req'range);
  signal   rx_dbts_req_rr     : std_logic_vector(rx_dbts_req'range);
  signal   rx_dbts_req_rrr    : std_logic_vector(rx_dbts_req'range);
  signal   rx_dbts_inh        : std_logic_vector(rx_dbts_req'range);
  signal   rx_dbts_enable     : std_logic;              -- de-bitstuffing
  signal   rx_ccnt            : unsigned(7 downto 0);   -- Rx cycle count
  signal   rx_ccnt_sval       : unsigned(7 downto 0);
  signal   rx_ccnt_set        : std_logic;
  signal   rx_ccnt_inh        : std_logic := '0';

  signal   rx_binf_cnt        : unsigned(4 downto 0);             -- Bytes in frame counter
  signal   rx_binf_clr        : std_logic;
  signal   rx_binf_inc        : std_logic;

  type state_rcv_type is (SR_IDLE, SR_WAIT_BEGIN, SR_GET_FRAME);
  signal   cstate_rx          : state_rcv_type;
  signal   nstate_rx          : state_rcv_type;

  signal   rx_byte_c          : std_logic_vector(7 downto 0);
  signal   rx_byte_c_rbo      : std_logic_vector(7 downto 0);
  signal   rx_balign          : std_logic;
  signal   rx_crc_cap         : std_logic;
  signal   rx_crc_clr         : std_logic;
  signal   rx_enable          : std_logic := '0';
  signal   rx_enable_clr      : std_logic;
  signal   rx_frame_rdy_flag  : std_logic := '0';
  signal   rx_svl_cmd         : std_logic := '0';
  signal   rx_svl_cmd_flag    : std_logic := '0';

  --
  -- Dual-clock domain support
  -- signal   tx_clk           : std_logic;  -- FIXME
  -- signal   tx_rst           : std_logic;
  -- signal   rx_clk           : std_logic;
  -- signal   rx_rst           : std_logic;

  signal    hdlc_tx_start_syntx    : std_logic_vector(2 downto 0);
  signal    hdlc_tx_start_re_synrx : std_logic_vector(1 downto 0);
  attribute tig                    : string;
  attribute async_reg              : string;
  attribute shift_extract          : string;
  attribute tig of hdlc_tx_start_re : signal is "yes";
  attribute tig of hdlc_tx_start    : signal is "yes";
  attribute async_reg of hdlc_tx_start_re : signal is "true";
  attribute async_reg of hdlc_tx_start    : signal is "true";
  attribute shift_extract of hdlc_tx_start_re_synrx : signal is "no";
  attribute shift_extract of hdlc_tx_start_syntx : signal is "no";

begin
  -- rx_clk <= clk;   -- FIXME: (temporary) port to two clock domains for
  -- tx_clk <= clk;   --          use in current G-RORC
  -- rx_rst <= rst;
  -- tx_rst <= rst;

  --------------------------------------------------------------------------------
  -- Software control interface
  --------------------------------------------------------------------------------

  --
  -- Request logic
  --   There's only support for a single frame in flight (for resource reasons)
  -- 
  --   Rising edge on hdlc_tx_start triggers transmission of the Tx payload and
  --   enables reception of _one_ reply frame
  process(rx_clk)
  begin
    if rising_edge(rx_clk) then

      if rx_rst = '1' then
        rx_enable <= '0';
      elsif rx_enable_clr = '1' then
        rx_enable <= '0';
      elsif hdlc_tx_start_re_synrx(1) = '1' or hdlc_rx_start_re = '1' then
        rx_enable <= '1';
      end if;

      -- Syncing to rx clock domain
      --   FIXME: use proper synchronizer entity for that...
      hdlc_tx_start_re_synrx(1) <= hdlc_tx_start_re_synrx(0);
      hdlc_tx_start_re_synrx(0) <= hdlc_tx_start_re;


      hdlc_rx_start_r <= hdlc_rx_start;
    end if;
  end process;

  hdlc_rx_start_re <= hdlc_rx_start and (not hdlc_rx_start_r);

  process(tx_clk)
  begin
    if rising_edge(tx_clk) then

      if tx_rst = '1' then
        tx_request <= '0';
      elsif tx_done = '1' then
        tx_request <= '0';
      elsif hdlc_tx_start_re = '1' then
        tx_request <= '1';
        tx_mode    <= hdlc_tx_mode;
        tx_addr    <= hdlc_tx_addr;
      end if;

      -- Syncing in start signal from PCIe
      --   FIXME: use proper synchronizer entity for that...
      hdlc_tx_start_syntx(2) <= hdlc_tx_start_syntx(1);
      hdlc_tx_start_syntx(1) <= hdlc_tx_start_syntx(0);
      hdlc_tx_start_syntx(0) <= hdlc_tx_start;

      -- hdlc_tx_start_r <= hdlc_tx_start;
    end if;
  end process;

  -- FIXME: edge detection on start signal from PCIe (on sync'd in signals)
  -- hdlc_tx_start_re <= hdlc_tx_start and (not hdlc_tx_start_r);
  hdlc_tx_start_re <= hdlc_tx_start_syntx(2) and (not hdlc_tx_start_syntx(1));



  --------------------------------------------------------------------------------
  -- HDLC reception
  --------------------------------------------------------------------------------

  --
  -- Rx de-bitstuffing detection
  --   successful de-bitstuffing detection requires grace interval for
  --   already checked bits to flow through shift register (realized by inhibit signals)
  rx_dbts_req(0) <= '1' when (rx_dbts_enable & rx_shift_reg(12 downto 7)) = "1111110" and
                              rx_dbts_inh(0) = '0'
                        else '0';
  rx_dbts_req(1) <= '1' when (rx_dbts_enable & rx_shift_reg(13 downto 8)) = "1111110" and
                              rx_dbts_inh(1) = '0'
                        else '0';

  process(rx_clk)
  begin
    if rising_edge(rx_clk) then
      rx_dbts_req_r    <= rx_dbts_req;
      rx_dbts_req_rr   <= rx_dbts_req_r;
      rx_dbts_req_rrr  <= rx_dbts_req_rr;
    end if;
  end process;

  rx_dbts_inh(0) <= rx_dbts_req_r(0) or rx_dbts_req_rr(0) or
                    rx_dbts_req_r(1) or rx_dbts_req_rr(1);
  rx_dbts_inh(1) <= rx_dbts_inh(0) or rx_dbts_req_rrr(0);

  rx_ccnt_inh <= '1' when rx_balign = '1' and rx_dbts_req /= "00" else '0';

  --
  -- Rx input shift register to shift in frame data
  --   need at least 3 byte plus 1 bit, so we can hold FCS + EOF and move Rx byte window
  process(rx_clk)
  begin
    if rising_edge(rx_clk) then

      if rx_dbts_req(0) = '1' then
        -- Stuffed bit is rx_shift_reg(7)
        rx_shift_reg <= rx_shift_reg(rx_shift_reg'high-1 downto 8) & rx_shift_reg(6 downto 0) & rx_2b;
      elsif rx_dbts_req(1) = '1' then
        -- Stuffed bit is rx_shift_reg(8)
        rx_shift_reg <= rx_shift_reg(rx_shift_reg'high-1 downto 9) & rx_shift_reg(7 downto 0) & rx_2b;
      else
        -- No stuffed bit detected, normal shift-in
        rx_shift_reg <= rx_shift_reg(rx_shift_reg'high-2 downto 0) & rx_2b;
      end if;
    end if;
  end process;

  --
  -- SOF/EOF and ABORT detection
  rx_sof_det(0) <= '1' when rx_shift_reg(7 downto 0) = HDLC_SOF else '0';   -- aligned even
  rx_sof_det(1) <= '1' when rx_shift_reg(8 downto 1) = HDLC_SOF else '0';   -- odd

  rx_abort_det(0) <= '1' when rx_shift_reg(6 downto 0) = "1111111" else '0';
  rx_abort_det(1) <= '1' when rx_shift_reg(7 downto 1) = "1111111" else '0';


  --
  -- Rx byte alignment flag
  process(rx_clk)
  begin
    if rising_edge(rx_clk) then
      if rx_rst = '1' then
        rx_balign <= '1';
      elsif rx_sof_det(0) = '1' then
        -- Frame marker detected on bits (7..0)
        rx_balign <= '1';
      elsif rx_sof_det(1) = '1' then
        -- Frame marker detected on bits (8..1)
        rx_balign <= '0';
      elsif rx_dbts_req /= "00" then
        -- De-bitstuff request toggles balign
        rx_balign <= not rx_balign;
      end if;
    end if;
  end process;


  --
  -- Rx cycle counter and byte-in-frame counter
  process(rx_clk)
  begin
    if rising_edge(rx_clk) then
      if rx_ccnt_set = '1' then
        rx_ccnt <= rx_ccnt_sval;
      elsif rx_ccnt_inh = '0' and rx_ccnt /= to_unsigned(0, rx_ccnt'length) then
        rx_ccnt <= rx_ccnt - 1;
      end if;

      if rx_binf_clr = '1' then
        rx_binf_cnt <= (others => '0');
      elsif rx_binf_inc = '1' then
        rx_binf_cnt <= rx_binf_cnt + 1;
      end if;
    end if;
  end process;

  rx_byte_c <= rx_shift_reg(23 downto 16) when rx_balign = '1' else
               rx_shift_reg(24 downto 17);

  gen_rbo0: for i in rx_byte_c'range generate
    rx_byte_c_rbo(i) <= rx_byte_c(rx_byte_c'high - i);    -- reverse bit order
  end generate;

  rx_hdlc_fcs_c <= rx_shift_reg(23 downto 8) when rx_balign = '1' else
                   rx_shift_reg(24 downto 9);

  gen_rbo1: for i in rx_hdlc_fcs_c'range generate
    rx_hdlc_fcs_c_rbo(i) <= rx_hdlc_fcs_c(rx_hdlc_fcs_c'high - i);  -- reverse bit order
  end generate;

  --
  -- Rx FSM
  process(rx_clk)
  begin
    if rising_edge(rx_clk) then
      if rx_rst = '1' then
        cstate_rx <= SR_IDLE;
      else
        cstate_rx <= nstate_rx;
      end if;
    end if;
  end process;

  process(cstate_rx, rx_sof_det, rx_ccnt, rx_ccnt_inh, rx_binf_cnt, rx_enable, rx_abort_det)
  begin
    nstate_rx <= cstate_rx;
    rx_dbts_enable <= '0';
    rx_ccnt_set <= '0';
    rx_ccnt_sval <= (others => '0');
    rx_binf_inc  <= '0';
    rx_binf_clr <= '0';
    rx_crc_cap <= '0';
    rx_crc_clr <= '1';
    rx_enable_clr <= '0';
    rx_svl_cmd <= '0';

    case cstate_rx is
      when SR_IDLE =>
        rx_crc_clr <= '0';

        if rx_enable = '1' and rx_sof_det /= "00" then
          -- Start of frame detected,
          --   wait until valid byte in upper byte of shift reg
          nstate_rx <= SR_WAIT_BEGIN;
          rx_binf_clr <= '1';
          rx_ccnt_sval <= to_unsigned(2 * 4 - 2, rx_ccnt'length);
          rx_ccnt_set  <= '1';
        end if;

      when SR_WAIT_BEGIN =>

        if rx_abort_det /= "00" then
          nstate_rx <= SR_IDLE;
        elsif rx_ccnt = to_unsigned(0, rx_ccnt'length) then
          nstate_rx <= SR_GET_FRAME;
          rx_dbts_enable <= '1';
        end if;

      when SR_GET_FRAME =>
        rx_dbts_enable <= '1';

        if rx_abort_det /= "00" then
          nstate_rx <= SR_IDLE;

        elsif rx_sof_det /= "00" then  -- EOF for this frame 
          rx_crc_cap <= '1';
          -- Only single frame operation supported
          --   go to idle state and disable rx_enabled
          nstate_rx <= SR_IDLE;
          rx_enable_clr <= '1';

          if rx_binf_cnt = to_unsigned(4, rx_binf_cnt'length) then
            -- Supervisory level commands contain ADDR (8bit), CTRL (8bit), CRC (16bit),
            --   i.e. 4 byte
            rx_svl_cmd <= '1';
          end if;

        elsif rx_ccnt = to_unsigned(0, rx_ccnt'length) and rx_ccnt_inh = '0' then
          rx_ccnt_sval <= to_unsigned(4 - 1, rx_ccnt'length);
          rx_ccnt_set <= '1';
          rx_binf_inc <= '1';
        end if;

        if rx_binf_cnt /= to_unsigned(0, rx_binf_cnt'length) then
          rx_crc_clr <= '0'; 
        end if;

    end case;
  end process;

  --
  -- Rx frame sequence number
  --   only lower 3 bit required for control field, keep larger for debugging
  process(rx_clk)
  begin
    if rising_edge(rx_clk) then
      if rx_rst = '1' then
        rx_seq_nr <= to_unsigned(0, rx_seq_nr'length);
      elsif rx_enable_clr = '1' and rx_svl_cmd = '0' then
        rx_seq_nr <= rx_seq_nr + C_RX_SEQ_NR_INC;
      end if;
    end if;
  end process;

  hdlc_rx_seq_nr <= std_logic_vector(rx_seq_nr);

  --
  -- Capture Rx frame data
  process(rx_clk)
  begin
    if rising_edge(rx_clk) then
      if rx_binf_inc = '1' then
        case to_integer(rx_binf_cnt) is
          -- 0 -> SOF
          when  1 => rx_hdlc_addr <= rx_byte_c_rbo;
          when  2 => rx_hdlc_ctrl <= rx_byte_c_rbo;
          --
          when  3 => rx_hdlc_data( 7 downto  0) <= rx_byte_c_rbo;
          when  4 => rx_hdlc_data(15 downto  8) <= rx_byte_c_rbo;
          when  5 => rx_hdlc_data(23 downto 16) <= rx_byte_c_rbo;
          when  6 => rx_hdlc_data(31 downto 24) <= rx_byte_c_rbo;
          when  7 => rx_hdlc_data(39 downto 32) <= rx_byte_c_rbo;
          when  8 => rx_hdlc_data(47 downto 40) <= rx_byte_c_rbo;
          when  9 => rx_hdlc_data(55 downto 48) <= rx_byte_c_rbo;
          when 10 => rx_hdlc_data(63 downto 56) <= rx_byte_c_rbo;
          --
          when others => null;
        end case;
      end if;
    end if;
  end process;

  --
  -- Rx CRC
  process(rx_clk)
  begin
    if rising_edge(rx_clk) then
      --
      -- Recalculate CRC on received data
      if rx_crc_clr = '1' then
        rx_hdlc_fcs_recalc <= x"ffff";
      elsif rx_binf_inc = '1' then
        rx_hdlc_fcs_recalc <= next_crc16_d8(rx_hdlc_fcs_recalc, rx_byte_c_rbo);
      end if;

      --
      -- Capture shipped FCS field
      if rx_crc_cap = '1' then
        rx_hdlc_fcs <= rx_hdlc_fcs_c_rbo;
      end if;
    end if;
  end process;


  --
  -- Rx frame status flags
  process(rx_clk)
  begin
    if rising_edge(rx_clk) then
      if rx_crc_cap = '1' then
        rx_frame_rdy_flag <= '1';
      elsif rx_enable = '1' then
        rx_frame_rdy_flag <= '0';
      end if;

      if rx_crc_cap = '1' and rx_svl_cmd = '1' then
        rx_svl_cmd_flag <= '1';
      elsif rx_enable = '1' then
        rx_svl_cmd_flag <= '0';
      end if;
    end if;
  end process;

  --
  -- Output mappings
  hdlc_rx_rdy     <= rx_frame_rdy_flag;
  hdlc_rx_svl_cmd <= rx_svl_cmd_flag;
  hdlc_rx_payload <= rx_hdlc_data(55 downto 48) &
                     rx_hdlc_data(63 downto 56) &
                     rx_hdlc_data(39 downto 32) &
                     rx_hdlc_data(47 downto 40) &
                     rx_hdlc_data(23 downto 16) &
                     rx_hdlc_data(31 downto 24) &
                     rx_hdlc_data( 7 downto  0) &
                     rx_hdlc_data(15 downto  8);
  hdlc_rx_crc <= rx_hdlc_fcs_recalc & rx_hdlc_fcs;
  hdlc_rx_ctrl <= rx_hdlc_ctrl;
  hdlc_rx_addr <= rx_hdlc_addr;

  with cstate_rx select
    hdlc_rx_fsm <= std_logic_vector(to_unsigned(1, hdlc_rx_fsm'length)) when SR_IDLE,
                   std_logic_vector(to_unsigned(2, hdlc_rx_fsm'length)) when SR_WAIT_BEGIN,
                   std_logic_vector(to_unsigned(3, hdlc_rx_fsm'length)) when SR_GET_FRAME,
                   std_logic_vector(to_unsigned(0, hdlc_rx_fsm'length)) when others;



  --------------------------------------------------------------------------------
  -- HDLC sending
  --------------------------------------------------------------------------------

  --
  -- Tx FSM
  process(tx_clk)
  begin
    if rising_edge(tx_clk) then
      if tx_rst = '1' then
        cstate_tx <= SS_RESET;
      else
        cstate_tx <= nstate_tx;
      end if;
    end if;
  end process;

  process(cstate_tx, tx_request, tx_mode, tx_bcnt, tx_bcnt_inh)
  begin
    nstate_tx        <= cstate_tx;
    tx_shift_load_sel <= LOAD_NONE;
    tx_bcnt_dec       <= '0';
    tx_bcnt_sval      <= to_unsigned(0, tx_bcnt_sval'length);
    tx_done           <= '0';
    tx_crc_clr        <= '0';
    bts_enable        <= '0';
    tx_frame_cap      <= '0';
    tx_frame_shift8   <= '0';

    case cstate_tx is
      when SS_RESET =>
        -- Wake up by sending IDLE pattern (aka ABORT marker)
        tx_bcnt_sval <= to_unsigned(HDLC_ABORT_CYC - 1, tx_bcnt_sval'length);
        tx_shift_load_sel <= LOAD_ABORT;
        nstate_tx  <= SS_IDLE;

      when SS_IDLE =>
        tx_bcnt_dec <= '1';
        tx_crc_clr     <= '1';

        if tx_bcnt = to_unsigned(0, tx_bcnt'length) then
          if tx_request = '1' then
            -- Frame tx requested
            tx_frame_cap <= '1';
            tx_bcnt_sval <= to_unsigned(HDLC_SOF_CYC - 1, tx_bcnt_sval'length);
            tx_shift_load_sel <= LOAD_FRAME_SOF;
            nstate_tx <= SS_SHIFT_SOF;
          else
            -- Continue sending IDLE pattern (aka ABORT marker)
            tx_shift_load_sel <= LOAD_ABORT;
            tx_bcnt_sval <= to_unsigned(HDLC_ABORT_CYC - 1, tx_bcnt_sval'length);
          end if;
        end if;

      -- Send SOF marker
      when SS_SHIFT_SOF =>
        tx_bcnt_dec <= '1';

        if tx_bcnt = to_unsigned(0, tx_bcnt'length) then
          tx_frame_shift8 <= '1';
          if tx_mode /= TX_MODE_REGULAR then
            -- Transmit supervisory level commands (only HDLC address and control field)
            tx_bcnt_sval <= to_unsigned((tx_hdlc_addr'length + tx_hdlc_ctrl'length)/2 - 1, tx_bcnt_sval'length);
          else
            -- Transmit regular elink packet targeted for SCA
            tx_bcnt_sval <= to_unsigned((tx_hdlc_addr'length + tx_hdlc_ctrl'length + 
                                         tx_hdlc_payload'length + tx_hdlc_zero_pad'length)/2 - 1, tx_bcnt_sval'length);
          end if;
          tx_shift_load_sel <= LOAD_FRAME_LOW;
          nstate_tx <= SS_SHIFT_FRAME_LOW;
        end if;

      -- Send HDLC ADR, CTRL, PAYLOAD, ZERO_PAD
      -- Bit stuffing is allowed
      when SS_SHIFT_FRAME_LOW =>
        bts_enable <= '1';
        tx_bcnt_dec <= '1';

        if tx_bcnt = to_unsigned(0, tx_bcnt'length) then
          tx_bcnt_sval <= to_unsigned(tx_hdlc_fcs'length/2 - 1, tx_bcnt_sval'length);
          tx_frame_cap <= '1';
          tx_shift_load_sel <= LOAD_FRAME_FCS;
          nstate_tx <= SS_SHIFT_FCS;
        elsif tx_bcnt(1 downto 0) = "00" and tx_bcnt_inh = '0' then -- load new byte (we shift by 2bit/cyc)
          tx_shift_load_sel <= LOAD_FRAME_LOW;
          tx_frame_shift8 <= '1';
        end if;

      -- Send HDLC FCS
      -- Bit stuffing is allowed
      when SS_SHIFT_FCS =>
        bts_enable <= '1';
        tx_bcnt_dec <= '1';

        if tx_bcnt = to_unsigned(0, tx_bcnt'length) and tx_bcnt_inh = '0' then
          tx_bcnt_sval <= to_unsigned(HDLC_EOF_CYC - 1, tx_bcnt_sval'length);
          nstate_tx <= SS_SHIFT_EOF;
          tx_shift_load_sel <= LOAD_FRAME_EOF;
        elsif tx_bcnt(1 downto 0) = "00" then -- load new byte (we shift by 2bit/cyc)
          tx_shift_load_sel <= LOAD_FRAME_FCS;
        end if;

      -- Send EOF marker, then prep for abort pattern
      when SS_SHIFT_EOF =>
        tx_bcnt_dec <= '1';

        if tx_bcnt = to_unsigned(0, tx_bcnt'length) then
          tx_bcnt_sval <= to_unsigned(HDLC_ABORT_CYC - 1, tx_bcnt_sval'length);
          tx_shift_load_sel <= LOAD_ABORT;
          nstate_tx <= SS_IDLE;
          tx_done <= '1';
        end if;

    end case;
  end process;

  --
  -- Tx word counter
  process(tx_clk)
  begin
    if rising_edge(tx_clk) then
      if tx_bcnt_sval /= to_unsigned(0, tx_bcnt_sval'length) then
        tx_bcnt <= tx_bcnt_sval;
      elsif tx_bcnt_dec = '1' and tx_bcnt_inh = '0' and tx_bcnt /= to_unsigned(0, tx_bcnt'length) then
        tx_bcnt <= tx_bcnt - 1;
      end if;
    end if;
  end process;

  --
  -- Tx frame sequence number counter
  --   only 3 bit required for control field, keep larger for debugging reasons
  process(tx_clk)
  begin
    if rising_edge(tx_clk) then
      if tx_rst = '1' then
        tx_seq_nr <= (others => '0');
      elsif tx_done = '1' then
        tx_seq_nr <= tx_seq_nr + C_TX_SEQ_NR_INC;
      end if;
    end if;
  end process;

  hdlc_tx_seq_nr <= std_logic_vector(tx_seq_nr);

  --
  -- Tx HDLC frame assembly
  tx_hdlc_payload <= hdlc_tx_payload(55 downto 48) &
                     hdlc_tx_payload(63 downto 56) &
                     hdlc_tx_payload(39 downto 32) &
                     hdlc_tx_payload(47 downto 40) &
                     hdlc_tx_payload(23 downto 16) &
                     hdlc_tx_payload(31 downto 24) &
                     hdlc_tx_payload( 7 downto  0) &
                     hdlc_tx_payload(15 downto  8);

  with tx_mode select
    tx_hdlc_addr <= tx_addr         when TX_MODE_REGULAR,  -- I-frame (regular packet)
                    (others => '1') when others;           -- U-frames

  with tx_mode select
    tx_hdlc_ctrl <= TX_CTRL_SVL_CONNECT                              when TX_MODE_SVL_CONNECT,   -- U-frame SABM
                    TX_CTRL_SVL_RESET                                when TX_MODE_SVL_RESET,     -- U-frame RSET
                    TX_CTRL_SVL_TEST                                 when TX_MODE_SVL_TEST,      -- U-frame TEST
                    std_logic_vector(rx_seq_nr(2 downto 0)) & '0' &
                      std_logic_vector(tx_seq_nr(2 downto 0)) & '0'  when others;                -- I-frame (regular packet)

  tx_frame_low <= tx_hdlc_zero_pad & tx_hdlc_payload & tx_hdlc_ctrl & tx_hdlc_addr;

  --
  -- Tx HDLC frame shift register
  process(tx_clk)
  begin
    if rising_edge(tx_clk) then
      if tx_rst = '1' then
        tx_shift_reg <= "00" & HDLC_ABORT;
      elsif tx_shift_load = '1' then
        if bts_flag = '1' or bts_req /= "00" then
          if tx_shift = '1' then
            tx_shift_reg <= tx_shift_reg_c & tx_shift_reg(3 downto 2);
          else
            tx_shift_reg <= tx_shift_reg_c & tx_shift_reg(1 downto 0);
          end if;
        else
          tx_shift_reg <= "00" & tx_shift_reg_c;
        end if;
      elsif tx_shift = '1' then
        tx_shift_reg <= "00" & tx_shift_reg(tx_shift_reg'length - 1 downto 2);
      end if;
    end if;
  end process;

  tx_shift_load <= '1' when tx_shift_load_sel /= LOAD_NONE else '0';

  with tx_shift_load_sel select
    tx_shift_reg_c <= HDLC_SOF               when LOAD_FRAME_SOF,
                      tx_frame_c(7 downto 0) when LOAD_FRAME_LOW,
                      tx_hdlc_fcs_c          when LOAD_FRAME_FCS,
                      HDLC_EOF               when LOAD_FRAME_EOF,
                      HDLC_ABORT             when others;

  process(tx_clk)
  begin
    if rising_edge(tx_clk) then
      if tx_frame_cap = '1' then
        tx_frame_c <= tx_frame_low;
      elsif tx_frame_shift8 = '1' then
        tx_frame_c <= x"ff" & tx_frame_c(tx_frame_c'high downto 8);
      end if;
    end if;
  end process;

  with tx_frame_cap select
    tx_hdlc_fcs_c <= tx_hdlc_fcs( 7 downto 0) when '1',
                     tx_hdlc_fcs(15 downto 8) when others;

  tx_bcnt_inh <= '1' when (bts_flag = '1' and bts_req /= "00") else '0';  -- inhibit down-count when 2nd bitstuff detected
  tx_shift    <= '0' when (bts_flag = '0' and bts_req /= "00") else '1';  -- inhibit shift with 1st detected bitstuff,
                                                                          --   shift-out window is moved between 1st and 2nd bitstuff

  --
  -- Tx CRC calculation
  process(tx_clk)
  begin
    if rising_edge(tx_clk) then
      if tx_crc_clr = '1' then
        tx_hdlc_fcs <= x"ffff";
      elsif tx_frame_shift8 = '1' then
        tx_hdlc_fcs <= next_crc16_d8(tx_hdlc_fcs, tx_frame_c(7 downto 0));
      end if;
    end if;
  end process;

  --
  -- Tx bit stuffing
  process(tx_clk)
  begin
    if rising_edge(tx_clk) then
      bts_enable_r <= bts_enable;

      if tx_rst = '1' then
        bts_flag <= '0';
        tx_history <= (others => '0');
      else
        tx_history <= tx_2b_i & tx_history(tx_history'high downto 2);
        if bts_req /= "00" then
          bts_flag <= not bts_flag;
        end if;
      end if;
    end if;
  end process;

  --
  -- Determine whether bit-stuffing is required
  --   If 5 consecutive '1' are seen in output stream, an extra '0' needs to be forced into
  --   the data stream
  tx_history_4co <= '1' when tx_history(4 downto 0) = "11110" else '0';
  tx_history_5co <= '1' when tx_history(4 downto 0) = "11111" else '0';

  bts_req(0) <= ((bts_enable and bts_enable_r) or bts_enable_r) and tx_history_5co;
  bts_req(1) <= ((bts_enable and bts_enable_r) or bts_enable_r) and  
                (((not bts_flag) and tx_history_4co and tx_shift_reg(0)) or
                (bts_flag and tx_history_4co and tx_shift_reg(1)));

  bts_select <= bts_flag & bts_req;

  --
  -- Move output window to send depending on bit-stuffing flags
  with bts_select select
    tx_2b_i <= tx_shift_reg(0) & '0' when "001",
               tx_shift_reg(1) & '0' when "101",
               --
               '0' & tx_shift_reg(0) when "010",
               '0' & tx_shift_reg(1) when "110",
               --
               tx_shift_reg(1 downto 0) when "000",
               tx_shift_reg(2 downto 1) when "100",
               --
               "11" when others; -- "011" and "111" cannot occur;

  --
  -- Mapping of output signals
  hdlc_tx_rdy <= not tx_request;
  tx_2b       <= tx_2b_i(0) & tx_2b_i(1);

  with cstate_tx select
    hdlc_tx_fsm <= std_logic_vector(to_unsigned(1, hdlc_tx_fsm'length)) when SS_RESET,
                   std_logic_vector(to_unsigned(2, hdlc_tx_fsm'length)) when SS_IDLE,
                   std_logic_vector(to_unsigned(3, hdlc_tx_fsm'length)) when SS_SHIFT_SOF,
                   std_logic_vector(to_unsigned(4, hdlc_tx_fsm'length)) when SS_SHIFT_FRAME_LOW,
                   std_logic_vector(to_unsigned(5, hdlc_tx_fsm'length)) when SS_SHIFT_FCS,
                   std_logic_vector(to_unsigned(6, hdlc_tx_fsm'length)) when SS_SHIFT_EOF,
                   std_logic_vector(to_unsigned(0, hdlc_tx_fsm'length)) when others;

end arc;
