----------------------------------------------------------------------------------
-- HDLC core for GBT EC/IC communication
-- Frame encoding, bit-stuffing, CRC, ... has to be done in software
--
-- kirsch@fias.uni-frankfurt.de
-- Inspired by Kai Chen's HDLC core, see https://gitlab.cern.ch/kai/HDLC-ICEC
-- 
-- The Tx and Rx shift registers need to hold the expected maximum frame size
--   (worst case approx 20% extra are needed to hold bitstuffing bits).
--   The width is controlled by SHREG_WIDTH, which needs to be a multiple of 8 for
--   synchronization with the selected inter-frame fill pattern (0x7e).
--
-- GBT SCA frame, e.g., is (according to documentation) 112 bit including EOF/SOF.
--   Of those, 96 bit are egligble for bit-stuffing, therefore max expected frame
--   length is 16+96*1.2 ~= 131 bit
----------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

entity gbt_hdlc_sw is
  generic (
    -- Width of TX/RX shift register
    --   Includes bits introduced by bit-stuffing
    --   Needs to multiple of 8
    SHREG_WIDTH : natural := 256
  );
  port (
    tx_clk         : in  std_logic;
    rx_clk         : in  std_logic;
    tx_rst         : in  std_logic;
    rx_rst         : in  std_logic;
    tx_usr_rst     : in  std_logic;
    rx_usr_rst     : in  std_logic;
    --
    tx_data        : in  std_logic_vector(SHREG_WIDTH-1 downto 0);
    rx_data        : out std_logic_vector(SHREG_WIDTH-1 downto 0);
    --
    tx_start       : in  std_logic;
    tx_done        : out std_logic;
    tx_clr         : in  std_logic;   -- clear statistics, debug, ...
    tx_bitswap_sel : in  std_logic;
    tx_sw_seq_inc  : in  std_logic;   -- inc software seq cnt
    rx_ack         : in  std_logic;
    rx_clr         : in  std_logic;   -- clear statistics, debug, ...
    rx_done        : out std_logic;
    rx_bitswap_sel : in  std_logic;
    rx_sw_seq_inc  : in  std_logic;   -- inc software seq cnt
    --
    tx_sw_seq_cnt : out std_logic_vector(3 downto 0);   -- software bookkeeping / HDLC I-frame control
    rx_sw_seq_cnt : out std_logic_vector(3 downto 0);   -- software bookkeeping / HDLC I-frame control
    tx_seq_cnt    : out std_logic_vector(3 downto 0);   -- debug
    rx_seq_cnt    : out std_logic_vector(3 downto 0);   -- debug
    rx_sof_cnt    : out std_logic_vector(7 downto 0);   -- debug
    --
    rx_2b         : in  std_logic_vector(1 downto 0);
    tx_2b         : out std_logic_vector(1 downto 0)
  );
end gbt_hdlc_sw;

architecture arc of gbt_hdlc_sw is

  constant TX_CNT_WDT : positive := integer(ceil(log2(real(SHREG_WIDTH/2))));

  constant SOF : std_logic_vector(7 downto 0) := x"7e";       -- Start-of-frame
  alias    EOF : std_logic_vector(7 downto 0) is SOF;         -- End-of-frame
  constant NIF : std_logic_vector(6 downto 0) := "1111111";   -- Not-in-frame

  signal tx_rst_comb : std_logic;
  signal rx_rst_comb : std_logic;

  signal tx_data_i      : std_logic_vector(tx_data'range) := (others => '1');
  signal tx_start_r     : std_logic;
  signal tx_start_re    : std_logic;
  signal tx_start_req   : std_logic;
  signal tx_iffill      : std_logic_vector(7 downto 0);
  signal tx_2b_mux      : std_logic_vector(1 downto 0);
  signal tx_done_i      : std_logic;
  signal tx_done_r      : std_logic;
  signal tx_done_fe     : std_logic;
  signal tx_go          : std_logic;
  signal tx_cnt         : unsigned(TX_CNT_WDT-1 downto 0) := (others => '1');

  signal rx_2b_i        : std_logic_vector(1 downto 0);
  signal rx_data_i      : std_logic_vector(rx_data'range) := (others => '1');
  signal rx_sof_det     : std_logic_vector(1 downto 0);
  signal rx_nif_det     : std_logic;
  signal rx_idle        : std_logic := '0';
  signal rx_done_i      : std_logic;
  signal rx_ack_re      : std_logic;
  signal rx_ack_r       : std_logic;
  signal rx_capture     : std_logic_vector(1 downto 0);

begin
  assert ((SHREG_WIDTH mod 8) = 0 and SHREG_WIDTH /= 0) report "SHREG is not multiple of 8" severity error;

  --
  -- Input / output bitswapping
  --   These should be configured on whether core is used for EC or IC
  --      TX-EC = 1, RX-EC = 1
  --      TX-IC = 0, RX-IC = 0  ?verify!
  tx_2b   <= tx_2b_mux(1 downto 0) when tx_bitswap_sel = '0' else (tx_2b_mux(0) & tx_2b_mux(1));
  rx_2b_i <= rx_2b(1 downto 0)     when rx_bitswap_sel = '0' else (rx_2b(0) & rx_2b(1));

  --
  -- Trivial mappings
  tx_done    <= tx_done_i;

  --
  -- Reset from system or user-initiated
  rx_rst_comb <= rx_rst or rx_usr_rst;
  tx_rst_comb <= tx_rst or tx_usr_rst;

  --------------------------------------------------------------------------------
  -- TX side
  --------------------------------------------------------------------------------
  process(tx_clk)
  begin
    if rising_edge(tx_clk) then
      tx_start_r    <= tx_start;
      tx_done_r     <= tx_done_i;

      --
      -- Request transmission fo frame in buffer
      if tx_rst_comb = '1' or tx_done_fe = '1' then
        tx_start_req <= '0';
      elsif tx_start_re = '1' then
        tx_start_req <= '1';
      end if;

      --
      -- Tx counter
      if tx_rst_comb = '1' then
        tx_cnt <= (others => '0');
      elsif tx_go = '1' then
        tx_cnt <= to_unsigned(SHREG_WIDTH/2-1, tx_cnt'length);
      elsif tx_cnt /= to_unsigned(0, tx_cnt'length) then
        tx_cnt <= tx_cnt - 1;
      end if;

      --
      -- Shift new frame
      if tx_go = '1' then
        tx_data_i <= tx_data;
      else
        tx_data_i <= "11" & tx_data_i(tx_data_i'length-1 downto 2);
      end if;
    end if;
  end process;

  tx_start_re <= tx_start and (not tx_start_r);
  tx_done_fe <= (not tx_done_i) and tx_done_r;

  --
  -- Tx done flag
  tx_done_i <= '1' when tx_cnt = to_unsigned(0, tx_cnt'length) else '0';

  --
  -- Initiate frame send (synchronized with inter-frame-fill pattern)
  tx_go <= '1' when tx_start_req = '1' and tx_done_i = '1' and
                    tx_iffill(1 downto 0) = "01" else
           '0';

  --
  -- Tx inter-frame fill pattern
  process(tx_clk)
  begin
    if rising_edge(tx_clk) then
      if tx_rst_comb = '1' then
        tx_iffill <= '0' & NIF;
      else
        tx_iffill <= tx_iffill(1 downto 0) & tx_iffill(tx_iffill'length-1 downto 2);
      end if;
    end if;
  end process;

  --
  -- Mux between inter-frame fill pattern and data Tx shreg output
  tx_2b_mux <= tx_iffill(1 downto 0) when tx_done_i = '1' else tx_data_i(1 downto 0);



  --------------------------------------------------------------------------------
  -- RX side
  --------------------------------------------------------------------------------
  rx_done <= rx_done_i;

  --
  -- Detect Rx SOF (or EOF) 
  rx_sof_det(0) <= '1' when rx_data_i(7 downto 0) = SOF else '0';
  rx_sof_det(1) <= '1' when rx_data_i(8 downto 1) = SOF else '0';

  -- Capture frame
  rx_capture(0) <= rx_sof_det(0) and rx_idle and (not rx_done_i);
  rx_capture(1) <= rx_sof_det(1) and rx_idle and (not rx_done_i);

  --
  -- Detect patterns that are not allowed inside a frame, but are expected outside of a frame
  rx_nif_det    <= '1' when (rx_data_i(7 downto 1) = NIF or rx_data_i(6 downto 0) = NIF) else '0';

  --
  -- Rx shift register and flags
  process(rx_clk)
  begin
    if rising_edge(rx_clk) then
      rx_data_i <= rx_2b_i & rx_data_i(rx_data_i'length-1 downto 2);

      --
      -- Capture frame when SOF is detected, and wipe with acknowledge
      if rx_ack_r = '1' then
        rx_data <= (others => '1');
      elsif rx_capture(0) = '1' then
        rx_data <= rx_data_i;
      elsif rx_capture(1) = '1' then
        rx_data <= '1' & rx_data_i(rx_data_i'length-1 downto 1);
      end if;

      --
      -- rx_idle indicates that (at least) current frame has completely passed
      --   through shift register
      if rx_rst_comb = '1' or rx_nif_det = '1' then
        rx_idle <= '1';
      elsif rx_capture /= "00" then
        rx_idle <= '0';
      end if;

      --
      -- rx_done is raised once frame is ready, requires acknowledge by software
      --   to be reset
      if rx_rst_comb = '1' or rx_ack_re = '1' then
        rx_done_i <= '0';
      elsif rx_capture /= "00" then
        rx_done_i <= '1';
      end if;

      rx_ack_r <= rx_ack;
    end if;
  end process;

  rx_ack_re <= rx_ack and (not rx_ack_r);

  --
  -- Statistics and debugging
  b_stat : block
    signal tx_sw_seq_cnt_i : unsigned(tx_seq_cnt'range);
    signal rx_sw_seq_cnt_i : unsigned(rx_seq_cnt'range);
    signal rx_sw_seq_inc_r : std_logic;
    signal tx_sw_seq_inc_r : std_logic;
    signal tx_seq_cnt_i    : unsigned(tx_seq_cnt'range);
    signal rx_seq_cnt_i    : unsigned(rx_seq_cnt'range);
    signal rx_sof_cnt_i    : unsigned(rx_sof_cnt'range);
  begin
    process(tx_clk)
    begin
      if rising_edge(tx_clk) then
        tx_sw_seq_inc_r <= tx_sw_seq_inc;

        -- Software controlled sequence counter
        if tx_rst_comb = '1' or tx_clr = '1' then
          tx_sw_seq_cnt_i <= to_unsigned(0, tx_sw_seq_cnt_i'length);
        elsif tx_sw_seq_inc = '1' and tx_sw_seq_inc_r = '0' then
          tx_sw_seq_cnt_i <= tx_sw_seq_cnt_i + 1;
        end if;

        -- Hardware sequence counter (for comparison)
        if tx_rst_comb = '1' or tx_clr = '1' then
          tx_seq_cnt_i <= to_unsigned(0, tx_seq_cnt_i'length);
        elsif tx_start_re = '1' then
          tx_seq_cnt_i <= tx_seq_cnt_i + 1;
        end if;
      end if;
    end process;

    process(rx_clk)
    begin
      if rising_edge(rx_clk) then
        rx_sw_seq_inc_r <= rx_sw_seq_inc;

        -- Software controlled sequence counter
        if rx_rst_comb = '1' or rx_clr = '1' then
          rx_sw_seq_cnt_i <= to_unsigned(0, rx_sw_seq_cnt_i'length);
        elsif rx_sw_seq_inc = '1' and rx_sw_seq_inc_r = '0' then
          rx_sw_seq_cnt_i <= rx_sw_seq_cnt_i + 1;
        end if;

        -- Hardware sequence counter (for comparison)
        if rx_rst_comb = '1' or rx_clr = '1' then
          rx_seq_cnt_i <= to_unsigned(0, rx_seq_cnt_i'length);
        elsif rx_capture /= "00" then
          rx_seq_cnt_i <= rx_seq_cnt_i + 1;
        end if;
      end if;
    end process;

    process(rx_clk)
    begin
      if rising_edge(rx_clk) then
        -- Clipping counter for SOF detected
        if rx_rst_comb = '1' or rx_clr = '1' then
          rx_sof_cnt_i <= to_unsigned(0, rx_sof_cnt_i'length);
        elsif rx_sof_det /= "00" and
              rx_sof_cnt_i /= to_unsigned(2**rx_sof_cnt_i'length-1, rx_sof_cnt_i'length) then
          rx_sof_cnt_i <= rx_sof_cnt_i + 1;
        end if;
      end if;
    end process;

    tx_sw_seq_cnt <= std_logic_vector(tx_sw_seq_cnt_i);
    rx_sw_seq_cnt <= std_logic_vector(rx_sw_seq_cnt_i);
    tx_seq_cnt <= std_logic_vector(tx_seq_cnt_i);
    rx_seq_cnt <= std_logic_vector(rx_seq_cnt_i);
    rx_sof_cnt <= std_logic_vector(rx_sof_cnt_i);
  end block b_stat;

end arc;
