quietly set clk_period_ns 25
quietly set tb_path "sim:/tb"

#-------------------------------------------------------------------------------
# R/W routines for gbt_sca_verilog core access
#
proc sca_write_reg {sca_id addr data} {
  global clk_period_ns
  global tb_path
  force -freeze ${tb_path}/s${sca_id}_sca_addr  "$addr" 0
  force -freeze ${tb_path}/s${sca_id}_sca_datai "$data" 0
  force -freeze ${tb_path}/s${sca_id}_sca_wr 1 0
  run $clk_period_ns ns
  force -freeze ${tb_path}/s${sca_id}_sca_wr 0 0
  run [expr 10*$clk_period_ns] ns
}


proc sca_read_reg {sca_id addr {cycles 4}} {
  global clk_period_ns
  global tb_path
  force -freeze ${tb_path}/s${sca_id}_sca_addr $addr 0
  run [expr $cycles*$clk_period_ns] ns
  return [examine -hex ${tb_path}/s${sca_id}_sca_datao]
}


proc sca_write_packet {sca_id d0 d1} {
  # Send 'regular' SCA Elink packet
  #   d0 gets written to addr 2 of sca core (sca tr_id, data, length, command)
  #   d1 gets written to addr 3 of sca core (sca data)
  sca_write_reg $sca_id 2 $d0
  sca_write_reg $sca_id 3 $d1
  sca_write_reg $sca_id 0 00000010
  sca_write_reg $sca_id 0 00000000
  sca_read_reg $sca_id 1
}


proc sca_send_svl_cmd {sca_id {svl_cmd_id 0xa}} {
  # Trigger sending of supervisory level commands
  #   CONNECT 0xa
  #   RESET   0xb
  #   TEST    0xc
  sca_write_reg $sca_id 0 "0x[format %08X [expr (0000001 << 4) | ($svl_cmd_id & 0xf)]]"
  sca_write_reg $sca_id 0 "0x[format %08X [expr (0000000 << 4) | ($svl_cmd_id & 0xf)]]"
  sca_write_reg $sca_id 0 00000000
}

proc sca_send_svl_connect {sca_id} {sca_send_svl_cmd $sca_id 0xa}
proc sca_send_svl_reset   {sca_id} {sca_send_svl_cmd $sca_id 0xb}
proc sca_send_svl_test    {sca_id} {sca_send_svl_cmd $sca_id 0xc}


proc sca_read_packet {sca_id d0 d1 status} {
  # Read 'regular' packet
  upvar $d0 ld0
  upvar $d1 ld1
  upvar $status ls0
  # data[31..0] => d0
  set ld0 [sca_read_reg $sca_id 2]
  # data[63..32] => d1
  set ld1 [sca_read_reg $sca_id 3]
  # status => status
  set ls0 [sca_read_reg $sca_id 1]
}


proc sca_send_request {sca_id sca_trid sca_channel sca_length sca_command sca_data} {
  set d0 "0x[format %08X [expr (($sca_trid & 0xff) << 8) | \
               (($sca_channel & 0xff) << 0) | \
               (($sca_length & 0xff) << 24) | \
               (($sca_command & 0xff) << 16)]]"
  set d1 "0x[format %08X $sca_data]"
  #puts "$d0 $d1"
  sca_write_packet $sca_id $d0 $d1
}

proc hdlc_sw_send_frame {data {inst_path "sim:/tb_sw/DUT0/inst_0"}} {
  global clk_period_ns
  force -freeze ${inst_path}/tx_data 16#$data
  force -freeze ${inst_path}/tx_start 1 0, 0 ${clk_period_ns} ns
}


#-------------------------------------------------------------------------------
# Dump data received by cores to file for sim comparison
#
proc dump_sca_verilog_rx_data {fp {sca_core_instance_path "sim:/tb/DUT0"}} {
  # Dump the frame data received by the gbt_sca_verilog core to a text file
  set d [string tolower [examine -hex $sca_core_instance_path/cmp_gbt_rx_sca/cmp_ELinkControl/DATARX]]
  set s [string tolower [examine -hex $sca_core_instance_path/cmp_gbt_rx_sca/cmp_ELinkControl/STATUS]]
  set c [string tolower [examine -hex $sca_core_instance_path/cmp_gbt_rx_sca/cmp_ELinkControl/CTRL]]
  puts $fp "DATA=0x$d | STATUS=0x$s CTRL=0x$c"
}

proc dump_hdlc_light_rx_data {fp {hdlc_light_core_instance_path "sim:/tb/DUT0"}} {
  set svl_cmd [examine -binary $hdlc_light_core_instance_path/hdlc_rx_svl_cmd]
  puts "svl_cmd $svl_cmd"
  if {$svl_cmd == 0} {
    set d [string tolower [examine -hex $hdlc_light_core_instance_path/hdlc_rx_payload]]
    puts $fp "DATA=0x$d"
  } else {
    set c [string tolower [examine -unsigned $hdlc_light_core_instance_path/hdlc_rx_ctrl]]
    set a [string tolower [examine -unsigned $hdlc_light_core_instance_path/hdlc_rx_addr]]
    set z [string tolower [examine -unsigned $hdlc_light_core_instance_path/hdlc_rx_crc]]
    puts $fp "SVL  CTRL=0x[format %02X $c]  ADDR=0x[format %02X $a]  CRC=0x[format %04X $z]"
  }
}


#-------------------------------------------------------------------------------
# Debug routines
#
proc dump_dbg_hdlc_rx_header {fp} {
  puts $fp "
#---------------------------------------
# COLUMNS
#   frame_no
#   shift_reg (bin)
#   rx_in
#   rx_sof_det
#   rx_b_align
#   dbts_req
#   rx_binf_inc
#   rx_byte_c_rbo
#   ccnt_inh
#   ccnt
#   rx_state
#   shift_reg (hex)
#---------------------------------------
"
}

proc dump_dbg_hdlc_rx {fp {path "sim:/tb/DUT2"}} {
  set state [examine $path/cstate_rx]
  set rx_in [examine -radix binary $path/rx_2b]
  set rx_sof_det [examine -radix binary $path/rx_sof_det]
  set sr_bin [examine -radix binary $path/rx_shift_reg]
  set sr [examine -hex $path/rx_shift_reg]
  set rx_balign [examine $path/rx_balign]
  set rx_binf_inc [examine $path/rx_binf_inc]
  set rx_byte_c_rbo [examine -hex $path/rx_byte_c_rbo]
  set ccnt_inh [examine $path/rx_ccnt_inh]
  set ccnt [examine -radix unsigned $path/rx_ccnt]
  set dbts_req [examine -radix binary $path/rx_dbts_req]
  set seq_nr [examine -radix unsigned $path/rx_seq_nr]

  puts $fp "$seq_nr  |  $sr_bin   $rx_in   $rx_sof_det   $rx_balign   $dbts_req   $rx_binf_inc   0x$rx_byte_c_rbo  |  $ccnt_inh   $ccnt  $state  0x$sr"
}




proc dump_dbg_hdlc_tx {fp {path "sim:/tb/DUT2"}} {
  set seq_nr [examine -radix unsigned $path/tx_seq_nr]
  set state [examine $path/cstate_tx]
  set tx_out [examine -radix binary $path/tx_2b]
  set bts_enable [examine -radix binary $path/bts_enable]
  set bts_req [examine -radix binary $path/bts_req]
  set bts_select [examine -hex $path/bts_select]
  set bts_flag [examine -radix binary $path/bts_flag]
  set tx_history [examine -radix binary $path/tx_history]
  set sr_bin [examine -radix binary $path/tx_shift_reg]
  set sr_hex [examine -hex $path/tx_shift_reg]
  set sr_c_hex [examine -hex $path/tx_shift_reg_c]
  set tx_shift_load [examine -radix binary $path/tx_shift_load]
  set tx_bcnt [examine -radix unsigned $path/tx_bcnt]
  set tx_bcnt_inh [examine -radix binary $path/tx_bcnt_inh]

  puts $fp "$seq_nr  |  $sr_bin $tx_history  $tx_out   \[$bts_req   $bts_flag   0x$bts_select  $bts_enable\]   $tx_bcnt_inh   \[$tx_shift_load  0x$sr_c_hex\]  |  [format %2u $tx_bcnt]   $state  0x$sr_hex"
}
