-- ------------------------------------------------------------------------------
-- HDLC raw format frame dump
--   Receives serial data and searches for HDLC SOF/EOF markers
--   and dumps content to text file
--
--   Expects 2 bit per cycle input
--   Added field debug markings in line below frame are approximate only,
--     e.g. bit stuffing is not considered!
-- ------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_textio.all;
use std.textio.all;

entity gbt_hdlc_raw_dump is
  generic (
    filename : string := "sca_dump.txt"
  );
  port (
    clk             : in std_logic;
    rst             : in std_logic;
    ec_rx           : in std_logic_vector(1 downto 0)
  );
end gbt_hdlc_raw_dump;

architecture arc of gbt_hdlc_raw_dump is

  file thefile     : text open write_mode is filename;

  signal shift_in : std_logic_vector(17 downto 0);

  constant PATTERN_ABORT   : std_logic_vector(6 downto 0) := (others => '1');
  constant HDLC_ABORT      : std_logic_vector(7 downto 0) := x"7f";
  constant HDLC_SOF        : std_logic_vector(7 downto 0) := x"7e";

  signal sof_detected   : std_logic;
  signal frame_cont     : std_logic;
  signal in_frame       : std_logic := '0';

begin

  --
  -- Shift in values with each (r or f) edge of clock
  process(clk)
  begin
    if rising_edge(clk) then
      shift_in <= shift_in(16 downto 0) & ec_rx(0);
    elsif falling_edge(clk) then
      shift_in <= shift_in(16 downto 0) & ec_rx(1);
    end if;
  end process;

  sof_detected   <= '1' when shift_in(17 downto 10) = HDLC_SOF else '0';
  frame_cont     <= '1' when shift_in(9 downto 3) /= PATTERN_ABORT else '0';

  process(clk)
    variable theline : line;
  begin
    if rising_edge(clk) or falling_edge(clk) then
      if sof_detected = '1' then
        -- Frame start marker detected
        if in_frame = '0' then
          in_frame <= '1';

          write(theline, string'("Frame start @"));
          write(theline, time'image(now));
          writeline(thefile, theline);
          write(theline, shift_in(17 downto 10));

        -- Frame end marker detected
        else
          write(theline, shift_in(10));    
          writeline(thefile, theline);

          write(theline, string'("SOF    |ADDR   |CTRL   |TRID   |CHAN   |LEN    |CMD    |DATA"));
          writeline(thefile, theline);

          write(theline, string'("Frame end @"));
          write(theline, time'image(now));
          writeline(thefile, theline);
          write(theline, string'("-"));
          writeline(thefile, theline);

          -- Frame is done
          if frame_cont /= '1' then
            in_frame <= '0';

          -- New frame shares SOF with EOF
          else
            write(theline, string'("Frame (cont) start @"));
            write(theline, time'image(now));
            writeline(thefile, theline);
            write(theline, shift_in(17 downto 10));
          end if;
        end if;

      elsif in_frame = '1' then
        write(theline, shift_in(10));    

        -- Abort marker detected
        if frame_cont = '0' then
          write(theline, string'("  Abort @"));
          write(theline, time'image(now));
          writeline(thefile, theline);
          write(theline, string'("-"));
          writeline(thefile, theline);
          in_frame <= '0';
        end if;
      end if;
    end if;
  end process;

end arc;
