--==============================================================
-- file gbt_sca_w.vhd
----------------------------------------------------------------
-- author Filippo Costa (filippo.costa at cern.ch)
----------------------------------------------------------------
-- version 1.4
--
-- details
-- GBT w to hide some signals not needed
--
-- Last changes:
--  1.0 31-03-2016 PC first version
--  1.1 27-06-2016 PC adding PCIe interface and removing IC
--  1.2 30-06-2016 PC changed waitreq in ack and added ready signal
--  1.3 07-07-2016 PC added reading SM and some extra dbg signals
--  1.4 20-07-2016 PC removed READ SM
--==============================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.bus_pkg.all;

--=============================================================================
-- Entity declaration for gbt_w
--
-- Description:
-- GBT w to hide some signals not needed
--
-- Signals Input :
-- rst_i : active high rst
-- clk_i : 240 MHz clock ref serial lines
--
-- Signals Output :
-- 
--=============================================================================

entity gbt_sca_w is
  port (
    gbt_sca_rx_i : in  gbt_sc_rx_rec;
    gbt_sca_tx_i : in  gbt_sc_tx_rec;
    --================================
    -- SCA PCIe interface
    --================================
    sca_i        : in  sca_in_rec;
    sca_o        : out sca_out_rec;
    -- EC
    ecrx_i       : in  std_logic_vector(1 downto 0);
    ectx_o       : out std_logic_vector(1 downto 0)
    );
end gbt_sca_w;

architecture rtl of gbt_sca_w is
  signal s_write        : std_logic;
  signal s_wr_ack       : std_logic;
  signal s_rd_ack       : std_logic;
  signal s_ready        : std_logic;    -- signal used to communicate with UL.
  signal s_chipselect   : std_logic;
  signal s_readdata     : std_logic_vector(31 downto 0);
  -- reg
  signal s_write_r      : std_logic;
  signal s_chipselect_r : std_logic;
  signal s_data_r       : std_logic_vector(31 downto 0);
  signal s_add_r        : std_logic_vector(2 downto 0);
  -- dbg
  signal s_st           : std_logic_vector(3 downto 0);

begin
  --==============================================================
  -- =========== GBT RX ==============
  -- GBT RX clk domain
  -- Receives in input 2 EC bits and provid in output 32 bits data
  --==============================================================
  cmp_gbt_rx_sca : entity work.gbt_sca_verilog
    port map (
      clk              => gbt_sca_rx_i.clk,
      reset            => gbt_sca_rx_i.reset,
      s0_address       => sca_i.add,
      s0_writedata     => x"00000000",
      s0_chipselect    => '1',
      s0_write         => '0',
      s0_readdata      => sca_o.data,
      s0_waitrequest_n => open,
      s0_irq           => sca_o.irq,
      rx_sdr_i         => ecrx_i,
      tx_sdr_o         => open
      );

  --==============================================================
  -- =========== GBT TX ==============
  -- GBT TX clk domain
  -- Gets in input parallel data and sends it to EC field 2 bits
  --============================================================
  cmp_gbt_tx_sca : entity work.gbt_sca_verilog
    port map (
      clk              => gbt_sca_tx_i.clk,
      reset            => gbt_sca_tx_i.reset,
      s0_address       => s_add_r,
      s0_writedata     => s_data_r,
      s0_chipselect    => s_write_r,
      s0_write         => s_write_r,
      s0_readdata      => sca_o.data_dbg,  -- feedback dbg
      s0_waitrequest_n => s_wr_ack,
      s0_irq           => open,
      rx_sdr_i         => "11",
      tx_sdr_o         => ectx_o
      );

  --============================================
  -- data is latched when the write pulse is '1'
  --============================================
  p_reg_data : process(gbt_sca_tx_i.clk)
  begin
    if rising_edge(gbt_sca_tx_i.clk) then
      if sca_i.wr_p = '1' then
        s_data_r <= sca_i.data;
        s_add_r  <= sca_i.add;
      end if;
    end if;
  end process;

  --======================================
  -- FSM to write SCA protocol
  -- Wrapper to WHISHBONE INTERFACE
  --======================================
  b_write : block
    type t_states is (IDLE,
                      SEND_WRITE,
                      ST_WAIT
                      );
    signal s_cs : t_states;
    signal s_ns : t_states;

  begin
    p_sm : process(gbt_sca_tx_i.clk)
    begin
      if rising_edge(gbt_sca_tx_i.clk) then
        if gbt_sca_tx_i.reset = '1' then
          s_cs <= IDLE;
        else
          s_cs <= s_ns;
        end if;
      end if;
    end process p_sm;

    p_cmb : process(s_cs,
                    sca_i.wr_p,
                    s_wr_ack)
    begin
      -- DEFAULT
      s_ns    <= s_cs;
      s_write <= '0';
      s_ready <= '0';

      case s_cs is
        when IDLE =>
          s_st    <= x"f";
          s_ready <= '1';
          if sca_i.wr_p = '1' then
            s_ready <= '0';
            s_ns    <= SEND_WRITE;
          end if;

        when SEND_WRITE =>
          s_st    <= x"1";
          s_write <= '1';
          if s_wr_ack = '1' then
            s_write <= '0';
            s_ns    <= ST_WAIT;
          end if;

        when ST_WAIT =>
          s_st <= x"2";
          if s_wr_ack = '0' then
            s_ns <= IDLE;
          end if;
      end case;
    end process p_cmb;

--    p_reg : process(gbt_sca_tx_i.clk)
--    begin
--      if rising_edge(gbt_sca_tx_i.clk) then
        s_write_r   <= s_write;
        sca_o.ready <= s_ready;
        sca_o.st_sm <= s_st;
--      end if;
--    end process p_reg;
  end block b_write;

  b_cs : block
    signal clk_cs : std_logic;
    signal data_cs : std_logic_vector(31 downto 0);
    signal add_cs : std_logic_vector(2 downto 0);
    signal wr_p_cs : std_logic;
    signal rx_cs : std_logic_vector(1 downto 0);
    attribute keep : string;
    attribute keep of clk_cs : signal is "true"; 
    attribute keep of data_cs : signal is "true"; 
    attribute keep of add_cs : signal is "true"; 
    attribute keep of wr_p_cs : signal is "true"; 
    attribute keep of rx_cs : signal is "true"; 
  begin
    clk_cs <= gbt_sca_tx_i.clk;
    process (clk_cs)
    begin
      if rising_edge(clk_cs) then
        data_cs <= sca_i.data;
        add_cs <= sca_i.add;
        wr_p_cs <= sca_i.wr_p;
        rx_cs <= ecrx_i;
      end if;
    end process;
  end block;

end rtl;
