-- --------------------------------------------------------------------------------
--   HDLC light core testbench
-- --------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

use work.bus_pkg.all;

entity tb is
end tb;

architecture behavioral of tb is
  -- CONSTANTs
  constant c_CLK_PERIOD_40  : time      := 25 ns;  -- 40 MHz

  -- TB SIGNAL
  signal s_clk_40           : std_logic := '1';
  signal s_reset            : std_logic;

  -- SCA send with PiPPos interface
  signal s0_sca_datai       : std_logic_vector(31 downto 0) := (others => '0');
  signal s0_sca_datao       : std_logic_vector(31 downto 0) := (others => '0');
  signal s0_sca_addr        : std_logic_vector(2 downto 0)  := (others => '0');
  signal s0_sca_wr          : std_logic := '0';
  signal s0_sca_irq         : std_logic;
  signal s0_sca_irq_r       : std_logic;

  signal s1_sca_addr        : std_logic_vector(2 downto 0)  := "001";
  signal s1_sca_datao       : std_logic_vector(31 downto 0) := (others => '0');
  signal s1_sca_irq         : std_logic;
  signal s1_sca_irq_r       : std_logic;

  -- REC
  signal s0_to_sca : sca_in_rec;
  signal s1_to_sca : sca_in_rec;
  signal s0_from_sca : sca_out_rec;
  signal s1_from_sca : sca_out_rec;  

  -- IC
  signal s_sc_rx  : gbt_sc_rx_rec;
  signal s_sc_tx  : gbt_sc_tx_rec;
  signal s_sca_rx : std_logic_vector(1 downto 0);
  signal s_sca_tx : std_logic_vector(1 downto 0);

  signal gsl_clk              : std_logic;
  signal gsl_rst              : std_logic := '0';
  signal gsl_hdlc_frame_in    : std_logic_vector(63 downto 0) := (others => '0');
  signal gsl_hdlc_send        : std_logic := '0';
  signal gsl_hdlc_send_mode   : std_logic_vector(1 downto 0) := (others => '0');
  signal gsl_hdlc_send_addr   : std_logic_vector(7 downto 0) := (others => '0');
  signal gsl_hdlc_send_rdy    : std_logic := '0';
  signal gsl_hdlc_frame_out_0 : std_logic_vector(63 downto 0);
  signal gsl_hdlc_frame_out_1 : std_logic_vector(63 downto 0);
  signal gsl_hdlc_rcv_rdy_0   : std_logic := '0';
  signal gsl_hdlc_rcv_rdy_1   : std_logic := '0';
  signal gsl_hdlc_rcv_rdy_0_r : std_logic := '0';
  signal gsl_hdlc_rcv_rdy_1_r : std_logic := '0';
  signal gsl_ec_rx            : std_logic_vector(1 downto 0) := (others => '0');
  signal gsl_ec_tx            : std_logic_vector(1 downto 0);

begin

  s_sc_rx.clk <= s_clk_40;
  s_sc_tx.clk <= s_clk_40;

  s_sc_rx.reset <= s_reset;
  s_sc_tx.reset <= s_reset;

  -- ------------------------------------------------------------------------------
  -- Reference cores
  --   gbt_sca_verilog.v from microelectronics group
  -- ------------------------------------------------------------------------------

  --
  -- DUT0: gbt_sca_verilog reference
  s0_to_sca.data <= s0_sca_datai;
  s0_to_sca.add  <= s0_sca_addr;
  s0_to_sca.wr_p <= s0_sca_wr;

  s0_sca_datao   <= s0_from_sca.data;
  s0_sca_irq     <= s0_from_sca.irq;

  DUT0 : entity work.gbt_sca_w
    port map(
      gbt_sca_rx_i => s_sc_rx,
      gbt_sca_tx_i => s_sc_tx,
      --
      sca_i        => s0_to_sca,
      sca_o        => s0_from_sca,
      -- EC         
      ecrx_i       => s_sca_rx,
      ectx_o       => s_sca_tx
      );


  --
  -- Raw data dump of DUT1
  dump0 : entity work.gbt_hdlc_raw_dump
    generic map (filename => "DUT0_gbt_sca_verilog_tx_frames_serial.txt")
    port map (
      clk => s_sc_tx.clk,
      rst => s_sc_tx.reset,
      ec_rx => s_sca_tx
    );


  --
  -- DUT1: gbt_sca_verilog reference
  --    Receive only from DUT0
  s1_to_sca.data <= (others => '0');
  s1_to_sca.add  <= s1_sca_addr;
  s1_to_sca.wr_p <= '0';

  s1_sca_datao   <= s1_from_sca.data;
  s1_sca_irq     <= s1_from_sca.irq;

  DUT1 : entity work.gbt_sca_w
    port map(
      gbt_sca_rx_i => s_sc_rx,
      gbt_sca_tx_i => s_sc_tx,
      --            
      sca_i        => s1_to_sca,
      sca_o        => s1_from_sca,
      -- EC         
      ecrx_i       => s_sca_tx,
      ectx_o       => open
      );

  process(s_sc_tx.clk)
  begin
    if rising_edge(s_sc_tx.clk) then
      s0_sca_irq_r <= s0_sca_irq;
      s1_sca_irq_r <= s1_sca_irq;
    end if;
  end process;

  -- ------------------------------------------------------------------------------
  -- GBT HDLC core (for SCA) rewrite
  -- ------------------------------------------------------------------------------
  gsl_clk <= s_clk_40;
  gsl_rst <= s_reset;

  --
  -- write interface mockup (to stay compatible with PiPPos interfacing of gbt_sca_verilog)
  process(gsl_clk)
  begin
    if rising_edge(gsl_clk) then
      if s0_sca_addr = "010" and s0_sca_wr = '1' then
        gsl_hdlc_frame_in(31 downto 0) <= s0_sca_datai;
      end if;

      if s0_sca_addr = "011" and s0_sca_wr = '1' then
        gsl_hdlc_frame_in(63 downto 32) <= s0_sca_datai;
      end if;
    end if;
  end process;

  gsl_hdlc_send <= '1' when (s0_sca_wr = '1') and (s0_sca_addr = "000") and (s0_sca_datai(4) = '1')
                      else '0';

  with (s0_sca_addr & s0_sca_datai(3 downto 0)) select
    gsl_hdlc_send_mode <= "01" when ("000" & x"a"),
                          "10" when ("000" & x"b"),
                          "11" when ("000" & x"c"),
                          "00" when others;
  gsl_hdlc_send_addr <= x"ef"; 

  gsl_ec_rx <= s_sca_tx;
  s_sca_rx  <= gsl_ec_tx;

  --
  -- DUT2: hdlc_light core
  --   Tx/Rx to/from DUT1
  DUT2 : entity work.gbt_hdlc_light
    port map (
      rx_clk          => gsl_clk,
      rx_rst          => gsl_rst,
      tx_clk          => gsl_clk,
      tx_rst          => gsl_rst,

      hdlc_tx_payload => gsl_hdlc_frame_in,
      hdlc_tx_start   => gsl_hdlc_send,
      hdlc_rx_start   => '0',
      hdlc_tx_mode    => gsl_hdlc_send_mode,
      hdlc_tx_addr    => gsl_hdlc_send_addr,
      hdlc_tx_rdy     => gsl_hdlc_send_rdy,
      hdlc_rx_payload => gsl_hdlc_frame_out_0,
      hdlc_rx_rdy     => gsl_hdlc_rcv_rdy_0,
      hdlc_rx_svl_cmd => open,
      hdlc_rx_crc     => open,

      rx_2b           => gsl_ec_rx,
      tx_2b           => gsl_ec_tx
    );

  dump2 : entity work.gbt_hdlc_raw_dump
    generic map (filename => "DUT2_hdlc_light_tx_frames_serial.txt")
    port map (
      clk => gsl_clk,
      rst => gsl_rst,
      ec_rx => gsl_ec_tx
    );

  process(gsl_clk)
  begin
    if rising_edge(gsl_clk) then
      gsl_hdlc_rcv_rdy_0_r <= gsl_hdlc_rcv_rdy_0;
      gsl_hdlc_rcv_rdy_1_r <= gsl_hdlc_rcv_rdy_1;
    end if;
  end process;


  --
  -- DUT3: hdlc_light core
  --  Receive from DUT2 only
  DUT3 : entity work.gbt_hdlc_light
    port map (
      rx_clk          => gsl_clk,
      rx_rst          => gsl_rst,
      tx_clk          => gsl_clk,
      tx_rst          => gsl_rst,

      hdlc_tx_payload => (others => '0'),
      hdlc_tx_start   => gsl_hdlc_send,      -- trigger same as DUT3 to enable frame reception
      hdlc_rx_start   => '0',
      hdlc_tx_mode    => gsl_hdlc_send_mode,
      hdlc_tx_addr    => gsl_hdlc_send_addr,
      hdlc_tx_rdy     => open,
      hdlc_rx_payload => gsl_hdlc_frame_out_1,
      hdlc_rx_rdy     => gsl_hdlc_rcv_rdy_1,
      hdlc_rx_svl_cmd => open,
      hdlc_rx_crc     => open,

      rx_2b           => gsl_ec_tx,
      tx_2b           => open
    );


  --
  -- Clock & reset
  s_clk_40 <= not s_clk_40 after (c_CLK_PERIOD_40 / 2.0);

  process
  begin
    -- Assert Reset
    s_reset <= '1';
    wait for (c_CLK_PERIOD_40 * 100);
    s_reset <= '0';
    wait;
  end process;

end behavioral;
